#from __future__ import print_function
from datetime import datetime
#import yaml  #CiteSoftLocal will assume these are not available.
#import semantic_version #CiteSoftLocal will assume these are not available.
import re
import sys
import os

def eprint(*args, **kwargs):#Print to stderr
    print(*args, file=sys.stderr, **kwargs)

######GLOBAL VARIABLES SECTION######
citations_dict = {}
checkpoint_log_filename = "CiteSoftwareCheckpointsLog.txt"
consolidated_log_filename = "CiteSoftwareConsolidatedLog.txt"
uniqueIDs_log_filename  = "CiteSoftwareUniqueIDs.txt"
consolidated_CFF_filename = "CITATIONS.cff"
validate_on_fly = True#Flag.  If set to true, argument names will be checked in real time, and invalid argument names will result in a printed warning to the user
valid_optional_fields = ["version", "cite", "author", "doi", "url", "encoding", "misc"]
valid_required_fields = ['timestamp', 'unique_id', 'software_name']
######END OF GLOBAL VARIABLES SECTION######

#TODO: Make a global (in the module) for the absolute path to the log.  That way other modules can change the file_path through CiteSoft.set_log_abs_path(...) or by CiteSoft.log_abs_path = ...
#cwd = os...path....
#log_abs_path = cwd
# def set_log_abs_path(abs_path)
    # global log_abs_path
    # log_abs_path = abs_path

########## START SECTION OF CODE WHICH HAS FUNCTIONS THAT ARE TO BE USED AS DECORATORS ################
#The below functions are intended to be used as decorators.
#They are similar to the example "decorator_maker_with_arguments" at https://www.datacamp.com/community/tutorials/decorators-python
#To find the example, search for "decorator_maker_with_arguments" at the above link.
#function "inner" below is named 'decorator' in the above link and 'wrapper' below is named 'wrapper' in the above link.


#The function_call_cite function is intended to be used for creating a citation any time the wrapped function is called.
#This is expected to be the most common usage of CiteSoft.
#write_immediately means that the citations would be written immediately when encountered, not just stored in memory.
def function_call_cite(unique_id, software_name, write_immediately=True, **add_args):
    #the unique_id and the software_name are the only truly required args.
    #Optional args are: ["version", "cite", "author", "doi", "url", "encoding", "misc"]
    #Every arg must be a string.
    def inner(func):
        def wrapper(*args, **kwargs):
            add_citation(unique_id, software_name, write_immediately, **add_args)
            result = func(*args, **kwargs)
            return result
        return wrapper
    return inner

#function_call_cite used to be called module_call_cite, so the below line is for backwards compatibility.
module_call_cite = function_call_cite

#The after_call_compile_checkpoints_log function is intended to be used as a decorator and only adds citations *after* the function call is finished.
def after_call_compile_checkpoints_log(file_path="", empty_checkpoints=True):
    def inner(func):
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            compile_checkpoints_log(file_path=file_path, empty_checkpoints=empty_checkpoints)
            return result
        return wrapper
    return inner

#The after_call_compile_consolidated_log function is intended to be used as a decorator and makes the consolided log *after* the function call is finished.
def after_call_compile_consolidated_log(file_path="", compile_checkpoints=False):
    def inner(func):
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            compile_consolidated_log(file_path=file_path, compile_checkpoints=compile_checkpoints)
            return result
        return wrapper
    return inner

########## END SECTION OF CODE WHICH HAS FUNCTIONS THAT ARE TO BE USED AS DECORATORS ################

#The import_cite function is intended to be used at the top of a sofware module. This would be used when a module should be cited each time it is imported.
#write_immediately means that the citations would be written immediately when encountered, not just stored in memory.
def import_cite(unique_id, software_name, write_immediately=True, **kwargs):
    add_citation(unique_id, software_name, write_immediately, **kwargs)

#The add_citation function is the method which actually adds a citation.
#It's not intended to be called directly by the dev-user, but it could be.
def add_citation(unique_id, software_name, write_immediately=True, **kwargs):
    new_entry = {'unique_id' : unique_id, 'software_name' : software_name, 'timestamp' : get_timestamp()}
    for key in kwargs:
        if validate_on_fly:
            if not key in valid_optional_fields:
                eprint("Warning, " + key + " is not an officially supported argument name.  Use of alternative argument names is strongly discouraged.")
        if type(kwargs[key]) is not list:#Make sure single optional args are wrapped in a list
            kwargs[key] = [kwargs[key]]
        new_entry[key] = kwargs[key]
    if unique_id in citations_dict:#Check for duplicate entries(e.g. from calling the same function twice)
        citations_dict[unique_id] = compare_same_id(citations_dict[unique_id], new_entry)
    else:
        citations_dict[unique_id] = new_entry
    if write_immediately == True:
        compile_checkpoints_log()

#This function creates cff files for each entry based. Th CFF files name is the unique_id converted to a valid file name.
def create_cff(entry, file_path=""):
    if "CITATIONS" not in os.listdir():
        os.mkdir("./CITATIONS")
    import re
    unique_id_string = entry['unique_id']
    valid_file_name_string = re.sub('[^\w_.)( -]', '_', unique_id_string)#remove characters from unique_id_string that are disallowed in filenames strings. Replace with "_". #TODO: We should using an encoding rather than this simple replace, that way the unique_id conversion will be 1:1 and reversible. The reason I did not use an encoding (yet) is I wanted to make sure that the encoding used is easily replicated in all programming languages. It should not be python specific -- but maybe it would be okay if it is?
    cff_filename = valid_file_name_string + ".cff"
    if os.path.exists("./" + file_path + "/CITATIONS/IndividualCff/"):
        pass
    else:
       os.mkdir("./" + file_path + "/CITATIONS/IndividualCff/")
    with open("./" + file_path + "/CITATIONS/IndividualCff/" + cff_filename, 'w') as file:
        write_dict_to_cff(file, entry)
    #now write to the consolided CFF file (since we are appending, it will be created if it does not exist):
    with open("./" + file_path + "/CITATIONS/" + consolidated_CFF_filename, 'a') as file:
        file.write('---\n')
        write_dict_to_cff(file, entry)

#As of August 4th 2021, we will just make minimal CFF files with the minimal required fields:
# https://github.com/citation-file-format/citation-file-format/blob/main/examples/1.2.0/pass/minimal/CITATION.cff
def write_dict_to_cff(file, citation_dict):
    file.write('cff-version: 1.2.0\n')
    file.write('message:' + citation_dict['timestamp'] + '\n')
    file.write('title: ' + citation_dict['software_name'] + '\n')
    file.write('identifiers: \n')
    file.write(' - type: unique_id' + '\n')
    file.write('   value: '+ citation_dict['unique_id'] + '\n')
    if 'author' in citation_dict: #TODO: change this to have family-names and given-names. Also add orcid.
        file.write('authors: \n') 
        for authorName in citation_dict['author']:
            file.write(' - name: ' + authorName + '\n')
    other_cff_valid_fields = ["version", "doi", "url", "license"]
    for field in other_cff_valid_fields:
        if field in citation_dict:
            file.write(field + ": " + str(citation_dict[field][0]) + '\n') #Consider changing: currently CiteSoft makes all optional fields into a list, including the version number. So we are taking the first item in the list. 

def update_unique_IDs_file(file_path=""):
    if "CITATIONS" not in os.listdir():
        os.mkdir("./CITATIONS")
    written_unique_IDs = []
    if uniqueIDs_log_filename in os.listdir("./CITATIONS"): #check if the file exists already.
        #if it exists, grab the items from it and put them into written_unique_IDs.
        with open("./" + file_path + "/CITATIONS/" + uniqueIDs_log_filename, 'r') as file:
            written_unique_IDs = file.readlines()            
    #To consider: Do we need to specify what kind of linebreak characters are used? No, just use "\n" in the specifications.
    for unique_id_index, unique_id in enumerate(written_unique_IDs):
        written_unique_IDs[unique_id_index] = unique_id[:-1] #remove the \n at the end of each string.
    
    #remove the existing unique_IDs from the citations_dict.
    current_keys = list(citations_dict.keys())
    for dict_key in current_keys:
        current_unique_id = citations_dict[dict_key]['unique_id']
        if current_unique_id in written_unique_IDs:
            del citations_dict[current_unique_id]
    
    #write the remaining unique_id values.
    with open("./" + file_path + "/CITATIONS/" + uniqueIDs_log_filename, 'a') as file:
        for dict_key in citations_dict:
            current_unique_id = citations_dict[dict_key]['unique_id']
            if current_unique_id not in written_unique_IDs:
                file.write(current_unique_id + "\n")

#Normally, checkpoints are stored in a dictionary until they are exported.  
#The exporting happens either when requested to from add_citation or from compile_consolidated_log.
def compile_checkpoints_log(file_path="", empty_checkpoints=True):
    if "CITATIONS" not in os.listdir():
        os.mkdir("./CITATIONS")
    update_unique_IDs_file(file_path=file_path) #write any unique ideas
    with open("./" + file_path + "/CITATIONS/" + checkpoint_log_filename, 'a') as file:
        write_dict_to_output(file, citations_dict)
    for dict_key in citations_dict:
        create_cff(citations_dict[dict_key])
    if empty_checkpoints==True:
        citations_dict.clear()

def compile_consolidated_log(file_path="", compile_checkpoints=False):
    #CiteSoftLocal will use compile checkpoints since it cannot consolidate.
    compile_checkpoints_log()
    print("Warning: CiteSoftLocal cannot make a consolidated log. Citations have been exported to CiteSoftwareCheckpointsLog.txt")
    # ##BELOW IS THE ORIGINAL FUNCTION CODE3##
    # #compile_checkpoints =True must be used sparingly. Otherwise it can slow down performance when the checkpoints file gets large.
    # import copy
    # consolidated_dict = copy.deepcopy(citations_dict) #Make a copy of citations_dict before emptying it.
    # if compile_checkpoints == True:
        # compile_checkpoints_log()
    # #Grab the citations from the checkpoint log.        
    # if checkpoint_log_filename in os.listdir("./CITATIONS"): #check if the file exists already.
        # checkpoint_log_exists = True
    # else:
        # checkpoint_log_exists = False
    # if checkpoint_log_exists == True: #can only read file if it exists.
        # with open("./CITATIONS/" + checkpoint_log_filename, 'r') as file:
            # yaml_file_contents = yaml.safe_load_all(file)
            # for yaml_document in yaml_file_contents:
                # if yaml_document != None: #This is for 'blank' documents of "---" with nothing after that symbol.
                    # for citation_entry in yaml_document:
                        # id = citation_entry["unique_id"]
                        # if id in consolidated_dict:
                            # consolidated_dict[id] = compare_same_id(consolidated_dict[id], citation_entry)
                        # else:
                            # consolidated_dict[id] = citation_entry        
    # #Grab the citations from the consolidated log.    
    # if consolidated_log_filename in os.listdir("./CITATIONS"): #check if the file exists already.
        # consolidated_log_exists = True
    # else:
        # consolidated_log_exists = False
    # if consolidated_log_exists == True: #can only read file if it exists.
        # with open("./" + file_path + "/CITATIONS/" + consolidated_log_filename, "r") as file:
            # yaml_file_contents = yaml.safe_load_all(file)
            # for yaml_document in yaml_file_contents:
                # if yaml_document != None: #This is for 'blank' documents of "---" with nothing after that symbol.
                    # for citation_entry in yaml_document:
                        # id = citation_entry["unique_id"]
                        # if id in consolidated_dict:
                            # consolidated_dict[id] = compare_same_id(consolidated_dict[id], citation_entry)
                        # else:
                            # consolidated_dict[id] = citation_entry
    # with open("./CITATIONS/" + consolidated_log_filename, 'w') as file:
        # file.write('#Warning: CiteSoftwareConsolidatedLog.txt may not include all softwares used. It is the end-user’s responsibility to verify that the no software citations are missing relative to those recorded in the complete logfile, CiteSoftwareCheckpointsLog.txt . This verification is important to do when using two or more codes together for the first time.\n')
        # write_dict_to_output(file, consolidated_dict)

#Takes a dictionary, converts it to CiteSoft-compatible YAML, and writes it to file
def write_dict_to_output(file, dictionary):
    if len(dictionary) > 0:
        file.write('---\n')
        for key,dict in dictionary.items():
            file.write('-\n')
            for s in valid_required_fields:
                file.write('    ' + s + ': >-\n')
                file.write('    '*2 + dict[s] + '\n')
            for subkey in dict:
                if subkey not in valid_required_fields:
                    file.write('    ' + subkey + ':\n')
                    if type(dict[subkey]) is list:
                        for i in dict[subkey]:
                            file.write('    '*2 + '- >-\n')
                            file.write('    '*3 + i + '\n')
                    else:
                        file.write('    '*2 + '- >-\n')
                        file.write('    '*3 + dict[subkey] + '\n')

#Helper Functions

#Returns a string of the current time in the ISO 8601 format (YYYY-MM-DDThh:mm:ss).
def get_timestamp():
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%dT%H:%M:%S")
    return timestamp

#Compares two entries
#Returns : The entry which should be kept
def compare_same_id(old_entry, new_entry):
    return new_entry #CiteSoftLocal will not do comparisons. It will just return the new_entry.
    # ##BELOW IS THE ORIGINAL FUNCTION CODE##
    # old_has_version = "version" in old_entry
    # new_has_version = "version" in new_entry
    # if old_has_version and new_has_version:#If both entries have a version, compare them return and the return the greater(newer) version
        # old_ver_str = str(old_entry["version"][0])
        # new_ver_str = str(new_entry["version"][0])
        # #Initialize variables, assume strings are valid unless parsing fails
        # old_ver_semver_valid = True
        # new_ver_semver_valid = True
        # decimal_regex_str = "^[0-9]+\.[0-9]+$"#Match string with decimal point enclosed by at least one number on either side
        # if re.match(decimal_regex_str, old_ver_str):
            # old_ver_str += '.0'#To ensure semantic version parser handles a decimal value correctly
        # if re.match(decimal_regex_str, new_ver_str):
            # new_ver_str += '.0'#To ensure semantic version parser handles a decimal value correctly
        # try:
            # old_sv = semantic_version.Version(old_ver_str)
        # except ValueError:
            # old_ver_semver_valid = False
        # try:
            # new_sv = semantic_version.Version(new_ver_str)
        # except:
            # new_ver_semver_valid = False
        # if old_ver_semver_valid and new_ver_semver_valid:#If both entries have a valid SemVer version, keep the older one only if it's greater. Else, keep the newer one.
            # if old_sv > new_sv:
                # return old_entry
            # else:
                # return new_entry
        # elif old_ver_semver_valid:#If only the old entry has a valid SemVer version, keep it
            # return old_entry
        # elif new_ver_semver_valid:#If only the new entry has a valid SemVer version, keep it
            # return new_entry
        # else:
            # #Version comparison failed, use alphanumeric comparison
            # if old_ver_str > new_ver_str:
                # return old_entry
            # else:
                # return new_entry
    # elif old_has_version and not new_has_version:#If old entry has a version and the new entry doesn't, the entry with a version takes precedence
        # return old_entry
    # elif not old_has_version and new_has_version:#Likewise, if new entry has a version and the old entry doesn't, the entry with a version takes precedence
        # return new_entry
    # else:#If neither entry has a version, save the new entry
        # return new_entry



import sys
#sys.path.insert(0, '/mumpce/')
import PEUQSE.mumpce.Project as mumpceProject
import PEUQSE.mumpce.solution as mumpceSolution
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import cm #EAW 2020/01/07
import platform
if platform.system() == 'posix':
    matplotlib.use('Agg') #added by A. Savara June 29th, 2021.
import copy

class plotting_functions_class():
    def __init__(self, UserInput, samples = False): # The plots require samples.  Other plot settings are probably plotting-package specific.
        self.UserInput = UserInput
#        if not samples: #Deprecated.
#            print("Warning: Pass in the 'samples' keyword argument containing a numpy array of samples to plot.")
    
    def mu_and_cov_from_samples(self):
        mu = np.mean(self.samples, axis = 0)
        cov = np.cov(self.samples,rowvar=False)
        return mu, cov

    def mumpce_plots(self, model_parameter_info = {}, active_parameters = [], pairs_of_parameter_indices = [], posterior_mu_vector = 0, posterior_cov_matrix = 0, prior_mu_vector = 0, prior_cov_matrix = 0, contour_settings_custom = {'figure_name','fontsize','max_num_y_ticks','max_num_x_ticks','colormap_posterior_customized','colormap_prior_customized','contours_normalized','colorbars','axis_limits'}): # Pass empty keyword arguments for important parameters.  That way, warnings may be issued if they are not set.  There is not really a good default for these keyword arguments.  They depend entirely on the nature of the data being plotted.
        mumpceProjectObject = mumpceProject.Project() # A mumpce project object must be created.
        if len(model_parameter_info) == 0:
            print("Pass the 'model_parameter_info' argument to the mumpce_plots function.")
            model_parameter_info = np.array([{'parameter_number': 0, 'parameter_name': 'Parameter 0'},{'parameter_number': 1, 'parameter_name': 'Parameter 1'}])
        if len(active_parameters) == 0:
            print("Pass the 'active_parameters' argument to the mumpce_plots function.")
            active_parameters = np.array([0, 1]) 
        mumpceProjectObject.active_parameters = active_parameters
        #if len(pairs_of_parameter_indices) == 0:
        #    print("Pass the 'pairs_of_parameter_indices' argument to the mumpce_plots function.")
        #    mumpceProjectObject.pairsOfParameterIndices = [[0, 1]]
        #else:
        mumpceProjectObject.pairsOfParameterIndices = pairs_of_parameter_indices
        #if not posterior_mu_vector == 0:
        #    print("Pass the 'posterior_mu_vector' argument to the mumpce_plots function.")
        #    posterior_mu_vector = np.array([-0.58888733,1.1200355])
        #if not posterior_cov_matrix == 0:
        #    print("Pass the 'posterior_cov_matrix' argument to the mumpce_plots function.")
        #    posterior_cov_matrix = np.array([[ 0.0148872,-0.01894579],[-0.01894579,0.04284732]])
        #if not prior_mu_vector == 0:
        #    print("Pass the 'prior_mu_vector' argument to the mumpce_plots functions.")
        #    prior_cov_matrix = np.array([-0.98888733,0.8200355])
        #if not prior_cov_matrix == 0:
        #    prior_cov_matrix = 10*posterior_cov_matrix
        #    print("Pass the 'prior_cov_matrix' argument to the mumpce_plots functions.")
        mumpceProjectObject.model_parameter_info = model_parameter_info
        mumpceSolutionsObject = mumpceSolution.Solution(posterior_mu_vector, posterior_cov_matrix, initial_x=prior_mu_vector, initial_covariance=prior_cov_matrix)
        mumpceProjectObject.solution = mumpceSolutionsObject
        #if hasattr(UserInput,'figure_name'):
        #    contour_settings_custom['figure_name']=UserInput.figure_name
        #else:
        #    contour_settings_custom['figure_name']='mumpce_plots_04'
        #if hasattr(UserInput,'fontsize'):
        #    contour_settings_custom['fontsize'] = UserInput.fontsize        
        #else:
        #    contour_settings_custom['fontsize'] = 'auto'
        #if hasattr(UserInput,'max_num_y_ticks'):
        #    contour_settings_custom['max_num_y_ticks'] = UserInput.max_num_y_ticks
        #else:
        #    contour_settings_custom['max_num_y_ticks'] = 'auto'
        #if hasattr(UserInput,'max_num_x_ticks'):
        #    contour_settings_custom['max_num_x_ticks'] = UserInput.max_num_x_ticks
        #else:
        #    contour_settings_custom['max_num_x_ticks'] = 'auto'
        #if hasattr(UserInput,'colormap_posterior_customized'):
        #    contour_settings_custom['colormap_posterior_customized'] = UserInput.colormap_posterior_customized
        #else:
        #    contour_settings_custom['colormap_posterior_customized'] = "Oranges"
        #if hasattr(UserInput,'colormap_prior_customized'):
        #    contour_settings_custom['colormap_prior_customized'] = UserInput.colormap_prior_customized
        #else:
        #    contour_settings_custom['colormap_prior_customized'] = "Greens"
        #if hasattr(UserInput,'contours_normalized'):
        #    contour_settings_custom['contours_normalized'] = UserInput.contours_normalized
        #else:
        #    contour_settings_custom['contours_normalized'] = False
        #if hasattr(UserInput,'center_on'):
        #    contour_settings_custom['center_on'] = UserInput.center_on
        #else:
        #    contour_settings_custom['center_on'] = 'prior'
        #if hasattr(UserInput,'colorbars'):
        #    contour_settings_custom['colorbars'] = UserInput.colorbars
        #else:
        #    contour_settings_custom['colorbars'] = True
        mumpceProjectObject.plot_pdfs(mumpceProjectObject.pairsOfParameterIndices, contour_settings_custom = contour_settings_custom)


    def seaborn_scatterplot_matrix(self):
        #fig0, ax0 = plt.subplots()
        #if UserInput.verbose:
        #    print(np.mean(rate_tot_array,axis = 0))
        #ax0.plot(np.array(experiments_df['AcH - T']),np.mean(rate_tot_array,axis = 0), 'r')
        #ax0.plot(np.array(experiments_df['AcH - T']),np.array(experiments_df['AcHBackgroundSubtracted'])/2000,'g')
        #ax0.set_ylim([0.00, 0.025])
        #ax0.set_xlabel('T (K)')
        #ax0.set_ylabel(r'$rate (s^{-1})$')
        #ax0.legend(['model posterior', 'experiments'])
        #fig0.tight_layout()
        #fig0.savefig('tprposterior.png', dpi=220)
        #posterior_df = pd.DataFrame(samples,columns=[UserInput.parameterNamesAndMathTypeExpressionsDict[x] for x in UserInput.parameterNamesList])
        #pd.plotting.scatter_matrix(posterior_df)
        #plt.savefig('scatter_matrix_posterior.png',dpi=220)
        return

    def rate_tot_plot(self):
        return

def sampledParameterHistogramMaker(parameterSamples, parameterName, parameterNamesAndMathTypeExpressionsDict, sampledParameterFiguresDictionary, sampledParameterAxesDictionary, directory='', parameterInitialValue=None, parameterMAPValue=None, parameterMuAPValue=None, histogram_plot_settings={}):

        if len(histogram_plot_settings) == 0: #this means histogram_plot_settings was not provided, or was blank, in which case we will populate variables with defaults.
            histogram_plot_settings = copy.deepcopy(histogram_plot_settings) #first make a fresh copy so the original dictionary is not changed.
            histogram_plot_settings['histograms_as_density'] = False
            histogram_plot_settings['y_label'] = ''
            histogram_plot_settings['show_initial_value'] = True
            histogram_plot_settings['show_MAP_value'] = True
            histogram_plot_settings['show_muAP_value'] = True
            histogram_plot_settings['x_label_size'] = 16
            histogram_plot_settings['y_label_size'] = 16
            histogram_plot_settings['axis_font_size'] = 16
            histogram_plot_settings['dpi'] = 220
            histogram_plot_settings['vertical_linewidth'] = 1.5


        #The axis font size argument needs to be parsed into another form. #Code was made following answer by "binaryfunt" here https://stackoverflow.com/questions/3899980/how-to-change-the-font-size-on-a-matplotlib-plot
        axis_font = {'size':str(histogram_plot_settings['axis_font_size'])}
        
        #Now to get some more strings needed and then make the plot object.
        parameterIndex = list(parameterNamesAndMathTypeExpressionsDict).index(parameterName)
        sampledParameterFiguresDictionary[parameterName], sampledParameterAxesDictionary[parameterName] = plt.subplots()   #making plt objects    
        sampledParameterAxesDictionary[parameterName].hist(parameterSamples[:,parameterIndex], density=histogram_plot_settings['histograms_as_density']) #filling the object with data
        #setting the labels etc. and then exporting.

        if histogram_plot_settings['y_label'] == '': #will use defaults if ''        
            if histogram_plot_settings['histograms_as_density'] == False:
                sampledParameterAxesDictionary[parameterName].set_ylabel('Frequency', **axis_font)
            if histogram_plot_settings['histograms_as_density'] == True:
                sampledParameterAxesDictionary[parameterName].set_ylabel('Probability Density', **axis_font)
        sampledParameterAxesDictionary[parameterName].set_xlabel(parameterNamesAndMathTypeExpressionsDict[parameterName], **axis_font)

        # Add vertical lines at initial value, MAP, and, muAP if determined by UserInput
        # Make sure that the value passed has a value
        if histogram_plot_settings['show_initial_value'] and type(parameterInitialValue) != type(None):
            sampledParameterAxesDictionary[parameterName].axvline(x=parameterInitialValue, color='#00A5DF', linestyle='--', linewidth=histogram_plot_settings['vertical_linewidth'])
        if histogram_plot_settings['show_MAP_value'] and type(parameterMAPValue) != type(None):
            sampledParameterAxesDictionary[parameterName].axvline(x=parameterMAPValue, color='r', linestyle='--', linewidth=histogram_plot_settings['vertical_linewidth'])
        if histogram_plot_settings['show_muAP_value'] and type(parameterMuAPValue) != type(None):
            sampledParameterAxesDictionary[parameterName].axvline(x=parameterMuAPValue, color='k', linestyle='--', linewidth=histogram_plot_settings['vertical_linewidth'])


        sampledParameterAxesDictionary[parameterName].tick_params(axis='x', labelsize=histogram_plot_settings['x_label_size']) #TODO: make these labels sizes a setting that can be changed.
        sampledParameterAxesDictionary[parameterName].tick_params(axis='y', labelsize=histogram_plot_settings['y_label_size'])
        sampledParameterFiguresDictionary[parameterName].tight_layout()
        sampledParameterFiguresDictionary[parameterName].savefig(directory+'Histogram_sampling_'+str(parameterIndex)+'_'+parameterName+'.png', dpi=histogram_plot_settings['dpi'])

        
        
        #The above block makes code kind of like this in a dynamic fashion. Since we know how many we will need, a dictionary is used to avoid the need for 'exec' statements when making new parameters.
        # fig2, ax2 = plt.subplots()
        # ax2.hist(samples[:,1])
        # ax2.set_ylabel('frequency')
        # ax2.set_xlabel(r'$E_{a2}$')
        # fig2.tight_layout()
        # fig2.savefig('Ea2.png', dpi=220)

    #Make histograms for each parameter. Need to make some dictionaries where relevant objects will be stored.
def makeHistogramsForEachParameter(parameterSamples,parameterNamesAndMathTypeExpressionsDict, directory='', parameterInitialValue=None, parameterMAPValue=None, parameterMuAPValue=None, histogram_plot_settings={}):
    sampledParameterFiguresDictionary = copy.deepcopy(parameterNamesAndMathTypeExpressionsDict) #This must be a deep copy to perserve original.
    sampledParameterAxesDictionary = copy.deepcopy(parameterNamesAndMathTypeExpressionsDict) #This must be a deep copy to preserve original.
    #The below code was originally added by Eric Walker, then modified by Troy Gustke to add in the initialValue, MAPvalue, and mu_apvALUE in June 2021, and merged in by A. Savara.    
    for key, iv, mp, mup in zip(parameterNamesAndMathTypeExpressionsDict,parameterInitialValue,parameterMAPValue,parameterMuAPValue):
        parameterName = key
        initialValue = iv
        MAPValue = mp
        Mu_APValue = mup
        sampledParameterHistogramMaker(parameterSamples, parameterName, parameterNamesAndMathTypeExpressionsDict, sampledParameterFiguresDictionary, sampledParameterAxesDictionary, directory=directory, parameterInitialValue=initialValue, parameterMAPValue=MAPValue, parameterMuAPValue=Mu_APValue, histogram_plot_settings=histogram_plot_settings)        

def createSimulatedResponsesPlot(x_values, listOfYArrays, plot_settings={}, listOfYUncertaintiesArrays=[], showFigure=True, directory=''):
    exportFigure = True #This variable should be moved to an argument or something in plot_settings.
    #First put some defaults in if not already defined.
    x_values = np.array(x_values)
    if len(np.shape(x_values)) == 0: #This 2d line breaks regular arrays, but works when a 'zero length' array comes in (I don't understand how a zero length array can occur, but it has happened.)
        x_values = np.atleast_2d(x_values)
    if 'x_label' not in plot_settings: plot_settings['x_label'] = ''
    if 'y_label' not in plot_settings: plot_settings['y_label'] = ''
    if 'legendLabels' not in plot_settings: plot_settings['legendLabels'] = ''
    if 'figure_name' not in plot_settings: plot_settings['figure_name'] = 'simulatedResponsesPlot'
    if 'dpi' not in plot_settings: plot_settings['dpi']=220
    fig0, ax0 = plt.subplots()
    if 'fontdict' in plot_settings: 
        #There are various things that could be added to this fontdict. #https://www.tutorialexample.com/understand-matplotlib-fontdict-a-beginner-guide-matplotlib-tutorial/
        fontdict = plot_settings['fontdict']
        if 'size' in fontdict:
            ax0.tick_params(axis='x', labelsize=fontdict['size'])
            ax0.tick_params(axis='y', labelsize=fontdict['size'])
    else:
        fontdict = None #initializing with the matplotlib default
    ax0.set_xlabel(plot_settings['x_label'], fontdict=fontdict)
    ax0.set_ylabel(plot_settings['y_label'], fontdict=fontdict) #TODO: THis is not yet generalized (will be a function)
    #The error linewidth is going to be set to a thick value if we have a small number of points.
    if 'error_linewidth' in plot_settings: error_linewidth = plot_settings['error_linewidth'] # making a convenient local variable.
    if str(error_linewidth).lower() == 'auto':
        if len(x_values) == 1:
            error_linewidth = 20
        elif len(x_values) > 10:
            error_linewidth = 1
        elif len(x_values) <= 10:
            error_linewidth = 10
    if str(error_linewidth).lower() == 'none': error_linewidth = 0 #This will hide the rror bars if they are not desired.
    if 'y_range' in plot_settings: ax0.set_ylim(plot_settings['y_range'] )
    if len(listOfYArrays) == 3: #This generally means observed, mu_guess, map, in that order.
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):           
                ax0.plot(x_values,listOfYArrays[0],'g')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line..
                ax0.plot(x_values,listOfYArrays[1], '#00A5DF')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
                ax0.plot(x_values,listOfYArrays[2], 'r') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
                    
        if len(x_values) == 1: #This means there are single points, and we need to make symbols, by adding an "o".
            for seriesIndex in range(len(listOfYArrays)):           
                ax0.plot(x_values,listOfYArrays[0],'go')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='o', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], 'co')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[2], 'ro') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.
    elif len(listOfYArrays) == 4: #This generally means observed, mu_guess, map, mu_app
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values,listOfYArrays[0],'g')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], '#00A5DF')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                                             
                ax0.plot(x_values,listOfYArrays[2], 'r') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[3], 'k')  #k is black.
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[3], yerr=np.array(listOfYUncertaintiesArrays[3]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
        if len(x_values) == 1: #This means there are single points, and we need to make symbols, by adding an "o".
                ax0.plot(x_values,listOfYArrays[0],'go')
                if len(listOfYUncertaintiesArrays) >= 1: #If length is >=1, uncertainties for first data set.
                    ax0.errorbar(x_values, listOfYArrays[0], yerr=np.array(listOfYUncertaintiesArrays[0]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for experiments data series, just a line.
                ax0.plot(x_values,listOfYArrays[1], 'co')
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[1], yerr=np.array(listOfYUncertaintiesArrays[1]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[2], 'ro') 
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[2]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
                ax0.plot(x_values,listOfYArrays[3], 'ko')  #k is black. https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.plot.html#matplotlib.pyplot.plot
                if len(listOfYUncertaintiesArrays) > 1: #If length is >1, uncertainties for all data sets
                    ax0.errorbar(x_values, listOfYArrays[2], yerr=np.array(listOfYUncertaintiesArrays[3]).flatten(), fmt='.', barsabove=False, markersize=0, linewidth=error_linewidth, color="gray", ecolor="lightgray") #markersize=0 because we want no marker for this.                    
    else:
        if len(x_values) > 1: #This means there are enough data to make lines.        
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values[0],listOfYArrays[seriesIndex])
        if len(x_values) == 1: #This means there are single points, and we need to make symbols.
            for seriesIndex in range(len(listOfYArrays)):
                ax0.plot(x_values[0],listOfYArrays[seriesIndex], 'o')            
    if plot_settings['legend'] == True:
        ax0.legend(plot_settings['legendLabels']) #legends must be after plots are made.
    fig0.tight_layout()
    if exportFigure==True:
        fig0.savefig(directory + plot_settings['figure_name'] + '.png', dpi=plot_settings['dpi'])
    if showFigure==False:
        plt.close(fig0)
    return fig0

def makeTrisurfacePlot(xValues, yValues, zValues, exportFigure = True, figure_name="TrisurfacePlot", showFigure=True, directory=''):
    from mpl_toolkits.mplot3d import Axes3D #Although it does not look like this is called here, the InfoGain plots will fail without this line.
    fig1, ax1 =  plt.subplots(1)
    ax1 = plt.axes(projection ='3d')
    image = ax1.plot_trisurf(xValues,yValues, zValues)
    if exportFigure == True:
        fig1.savefig(directory + figure_name + '.png')
    if showFigure==False:
        plt.close(fig1)
    return fig1, ax1, image

def makeMeshGridSurfacePlot(XX, YY, ZZ,  plot_settings = {}, exportFigure = True, figure_name="MeshGridSurfacePlot", showFigure=True, directory=''):
    #TODO: plot_settings should be used for axis labels etc, like above.
    #TODO: create a UserInput variable named info_gain_plot_settings (like what the other cases have).
    from mpl_toolkits.mplot3d import Axes3D #I am not sure if this line is needed here, it might b.
    fig1,ax1 = plt.subplots(figsize=(5,5))
    #ax = fig.add_subplot(111, projection='3d')
    surf = ax1.pcolor(XX,YY,ZZ,cmap=matplotlib.cm.coolwarm)
    # ax1.set_xlabel('Temperature (K)')
    # ax1.set_ylabel(r'$p_A$')
    #ax.set_zlabel('Information Gain')
    #ax1.set_xticks(temperatures)
    #ax1.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
    ax1.set_title('Information Gain Surface')
    fig1.colorbar(surf, shrink=0.5, aspect=5)
    if exportFigure==True:
        fig1.savefig(directory + figure_name + '.png')
    if showFigure==False:
        plt.close(fig1)
    return fig1, ax1#, image
    

def density_scatter( x , y, ax = None, sort = True, bins = 20, **kwargs )   :
    """
    Scatter plot colored by 2d histogram
    Sampled from Guillaume post on https://stackoverflow.com/questions/20105364/how-can-i-make-a-scatter-plot-colored-by-density-in-matplotlib/53865762#53865762
    """
    from matplotlib import cm
    from matplotlib.colors import Normalize 
    from scipy.interpolate import interpn
    if ax is None:
        fig , ax = plt.subplots()
    data , x_e, y_e = np.histogram2d( x, y, bins = bins, density = True )
    z = interpn( ( 0.5*(x_e[1:] + x_e[:-1]) , 0.5*(y_e[1:]+y_e[:-1]) ) , data , np.vstack([x,y]).T , method = "splinef2d", bounds_error = False)
    #To be sure to plot all data
    z[np.where(np.isnan(z))] = 0.0

    # Sort the points by density, so that the densest points are plotted last
    if sort:
        idx = z.argsort()
        x, y, z = x[idx], y[idx], z[idx]

    ax.scatter( x, y, c=z, **kwargs )
    norm = Normalize(vmin = np.min(z), vmax = np.max(z))
    cbar = fig.colorbar(cm.ScalarMappable(norm = norm), ax=ax)
    cbar.ax.set_ylabel('Density')
    return fig, ax                   


def createScatterPlot(data_a, data_b, a_tuple, b_tuple, graphs_directory, plot_settings):
    """Generates and saves a scatter matrix plot.

    :param data_a: parameter points in first for loop (:type: pd.Series)
    :param data_b: parameter points in second for loop (:type: pd.Series)
    :param a_tuple: set of parameter names, MAP, muAP, and initial poitns (:type: tuple)
    :param b_tuple: set of parameter names, MAP, muAP, and initial poitns (:type: tuple)
    :param graphs_directory: path to graphs directory (:type: str)
    :param plot_settings: plot settings from User Input (:type: dict) https://github.com/AdityaSavara/PEUQSE/blob/master/PEUQSE/UserInput.py
    """
    import matplotlib.pyplot as plt
    point_plot_settings = (plot_settings['sampled_point_sizes'], plot_settings['sampled_point_transparency'])
    cross_plot_settings = (plot_settings['cross_marker_size'], plot_settings['cross_marker_transparency'])
    fig = plt.figure()
    # create a scatter plot of the posterior data between two parameters
    plt.scatter(data_a, data_b, s=point_plot_settings[0], alpha=point_plot_settings[1])
    # Allows the user to have no crosses if size is set to 0
    if cross_plot_settings[0] != 0:
        cross_size = cross_plot_settings[0]
        cross_transparency = cross_plot_settings[1]
        # creates the extra points for the MAP, muAP, and initial point denoted by crosses
        plt.scatter(a_tuple[2], b_tuple[2], s=cross_size, alpha=cross_transparency, c='r', marker='x') 
        plt.scatter(a_tuple[3], b_tuple[3], s=cross_size, alpha=cross_transparency, c='k', marker='x') 
        plt.scatter(a_tuple[4], b_tuple[4], s=cross_size, alpha=cross_transparency, c='#00A5DF', marker='x')
    # create labels and save the image to the graphs directory
    plt.xlabel(a_tuple[0], fontsize=plot_settings['fontsize'])
    plt.ylabel(b_tuple[0], fontsize=plot_settings['fontsize'])
    if plot_settings['max_num_x_ticks'] != 'auto' and isinstance(plot_settings['max_num_x_ticks'], int):
        plt.locator_params(axis='x', nbins=plot_settings['max_num_x_ticks'])
    if plot_settings['max_num_y_ticks'] != 'auto' and isinstance(plot_settings['max_num_y_ticks'], int):
        plt.locator_params(axis='y', nbins=plot_settings['max_num_y_ticks'])
    fig.savefig(graphs_directory+f'Scatter_{a_tuple[1]}_{b_tuple[1]}',dpi=plot_settings['dpi'])
    plt.close(fig)

def createScatterHeatMapPlot(data_a, data_b, a_tuple, b_tuple, graphs_directory, plot_settings):
    """Generates and saves a scatter heat map plot.

    :param data_a: parameter points in first for loop (:type: pd.Series)
    :param data_b: parameter points in second for loop (:type: pd.Series)
    :param a_tuple: set of parameter names, MAP, muAP, and initial poitns (:type: tuple)
    :param b_tuple: set of parameter names, MAP, muAP, and initial poitns (:type: tuple)
    :param graphs_directory: path to graphs directory (:type: str)
    :param plot_settings: plot settings from User Input (:type: dict) https://github.com/AdityaSavara/PEUQSE/blob/master/PEUQSE/UserInput.py
    """
    import matplotlib.pyplot as plt
    point_plot_settings = (plot_settings['sampled_point_sizes'], plot_settings['sampled_point_transparency'])
    cross_plot_settings = (plot_settings['cross_marker_size'], plot_settings['cross_marker_transparency'])
    # create a scatter plot of the posterior data between two parameters
    fig, ax = density_scatter(data_a, data_b, s=point_plot_settings[0], alpha=point_plot_settings[1])
    # Allows the user to have no crosses if size is set to 0
    if cross_plot_settings[0] != 0:
        cross_size = cross_plot_settings[0]
        cross_transparency = cross_plot_settings[1]
        # creates the extra points for the MAP, muAP, and initial point denoted by crosses
        ax.scatter(a_tuple[2], b_tuple[2], s=cross_size, alpha=cross_transparency, c='r', marker='x') 
        ax.scatter(a_tuple[3], b_tuple[3], s=cross_size, alpha=cross_transparency, c='k', marker='x') 
        ax.scatter(a_tuple[4], b_tuple[4], s=cross_size, alpha=cross_transparency, c='#00A5DF', marker='x')
    # create labels and save the image to the graphs directory
    ax.set_xlabel(a_tuple[0], fontsize=plot_settings['fontsize']) # , fontsize=plot_settings['fontsize']
    ax.set_ylabel(b_tuple[0], fontsize=plot_settings['fontsize']) # , fontsize=plot_settings['fontsize']
    if plot_settings['max_num_x_ticks'] != 'auto' and isinstance(plot_settings['max_num_x_ticks'], int):
        plt.locator_params(axis='x', nbins=plot_settings['max_num_x_ticks'])
    if plot_settings['max_num_y_ticks'] != 'auto' and isinstance(plot_settings['max_num_y_ticks'], int):
        plt.locator_params(axis='y', nbins=plot_settings['max_num_y_ticks'])
    fig.savefig(graphs_directory+f'Heat_Scatter_{a_tuple[1]}_{b_tuple[1]}',dpi=plot_settings['dpi'])
    plt.close(fig)

def createAutoCorrPlot(N, taus, param_name, param_symbol, graphs_directory):
    """
    Creates Integrated Autocorrelated Time Plots to show MCMC convergence.
    Convergence can be inferred when the AutoCorrelatedTime converges.

    :param N: Indices for the window size. (:type: np.array)
    :param taus: Autocorrelated Time values. (:type: np.array)
    :param param_name: Parameter name. (:type: str)
    :param param_symbol: Parameter symbol. Usually in unicode raw string. (:type: str)
    :param graphs_directory: Directory for storing graphs. (:type: str)
    """
    # create loglog plot
    fig = plt.figure()
    plt.loglog(N, taus, "o-", label='AutoCorr')
    # get current y boundaries
    ylim = plt.gca().get_ylim()
    # plot 50*tau heuristical line for possible convergence
    plt.plot(N, N / 50.0, "--k", label=r"$\tau = N/50$")
    # reset the y boundary
    plt.ylim(ylim)
    # create labels and legend
    plt.xlabel("number of samples, $N$")
    plt.ylabel(r"$\tau$ estimates")
    plt.title(f'Integrate Autocorrelation Time for {param_symbol}')
    plt.legend(fontsize=14)
    # save and close figure
    plt.savefig(graphs_directory+'AutoCorrelationPlot_'+param_name)
    plt.close(fig)

def createGewekePlot(z_scores, N, z_percents, param_name, param_symbol, graphs_directory):
    """
    Creates Gewekes indicies plot of entire post burn in samples
    and a plot for percentage of points that fall outside 1 std when compared to the 
    first 10% of the relative window size to the last 50%.

    :param z_scores: z scores for overall Geweke plot. (:type: np.array)
    :param N: Indices for the window size. (:type: np.array)
    :param z_percents: Percent of z scores outside 1 std for each window size. (:type: np.array)
    :param param_name: Parameter name. (:type: str)
    :param param_symbol: Parameter symbol. Usually in unicode raw string. (:type: str)
    :param graphs_directory: Directory for storing graphs. (:type: str)
    """
    # create 2 subplots for one figure
    fig, (ax1, ax2) = plt.subplots(1, 2, tight_layout=True, squeeze=True)
    plt.suptitle(f'Geweke Diagnostics for {param_symbol}')
    # plot Geweke plot on first graph
    ax1.scatter(*z_scores)
    # plot 1 std line. since this is abs of values only positive is needed.
    ax1.hlines([1], 0, round(z_scores[0][-1]), linestyles='dotted')
    # set labels and x boundaries
    ax1.set_xlabel('Last 50% Samples')
    ax1.set_ylabel('Z scores')
    ax1.set_xlim(0, round(z_scores[0][-1]))
    # plot Geweke percents
    ax2.plot(N, z_percents, "o-")
    # set labels, save figure, and close
    ax2.set_xlabel('Sample Indices')
    ax2.set_ylabel('Percent Outside '+u"\u00B1"+"1"+u"\u03C3")
    fig.savefig(graphs_directory+'GewekeDiagnostic_'+param_name)
    plt.close(fig)